module interp

data Expr = Val Int
          | Var String
          | Op (Int -> Int -> Int) Expr Expr
          | Lam String Expr
          | App Expr Expr

interp : List (String, Expr) -> Expr -> Maybe Expr
interp env (Val v) = Just (Val v)
interp env (Var x) = do val <- lookup x env
                        interp env val

interp env (Op op x y) = do x' <- interp env x
                            y' <- interp env y
                            runOp x' y'
   where runOp (Val x) (Val y) = Just (Val (op x y))
         runOp vx vy = Just (Op op vx vy)

interp env (App f a) = do f' <- interp env f
                          a' <- interp env a
                          runApp f' a'
   where runApp (Lam arg scope) a = interp ((arg, a) :: env) scope
         runApp f a = Just (App f a)

interp env (Lam v sc) = Just (Lam v sc)

test1 : Expr
test1 = App (Lam "a" (Op (+) (Var "a") (Var "a"))) (Val 21)

test2 : Expr
test2 = App (Lam "a" (Op (+) (Var "a") (Var "b"))) (Val 21)

test3 : Expr
test3 = App (Val 10) (Val 20)

